#ifndef MYSTRING__H__
#define MYSTRING__H__

// Copyright (C) 1998 Bruce Guenter <bruce.guenter@qcc.sk.ca>
//
// This package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

#include <string.h>
#include <iostream.h>

struct mystringrep
{
  unsigned length;
  unsigned references;
  char buf[1];

  void attach();
  void detach();
  
  static mystringrep* alloc(unsigned);
  static mystringrep* dup(const char*, unsigned);
  static mystringrep* dup(const char*, unsigned,
			  const char*, unsigned);
};

class mystring
{
private:
  mystringrep* rep;

protected:
  void dupnil();
  void dup(mystringrep*);
  void dup(const char*);
  void dup(const char*, size_t);
  void dup(const char*, const char*);
  void dup(const char*, const char*, const char*);
  void append(const char*, size_t);

public:
  mystring() { dupnil(); }
  mystring(const char* x) { dup(x); }
  mystring(const char* str, size_t len) { dup(str, len); }
  mystring(const mystring& that) { dup(that.rep); }
  mystring(const char* str1, const char* str2) { dup(str1, str2); }
  mystring(const char* str1, const char* str2, const char* str3)
    {
      dup(str1, str2, str3);
    }
  ~mystring();

  const char* c_str() const { return rep->buf; }

  bool operator!() const { return empty(); }

  char operator[](size_t i) const { return rep->buf[i]; }
  
  size_t length() const { return rep->length; }

  bool empty() const { return rep->length == 0; }
  
  int operator!=(const char* in) const
  {
    if(in == rep->buf)
      return 0;
    return (!in || !in[0]) ? rep->length!=0 : strcmp(rep->buf, in);
  }
  
  bool operator==(const char* in) const
  {
    return !operator!=(in);
  }

  int operator!=(const mystring& in) const
    {
      if(in.rep == rep)
	return 0;
      return strcmp(rep->buf, in.rep->buf);
    }
  bool operator==(const mystring& in) const
    {
      return !operator!=(in);
    }
  
  mystring operator=(const char*);
  mystring operator=(const mystring&);
  
  mystring subst(char from, char to) const;
  
  mystring lower() const;
  mystring upper() const;

  int find(char, size_t = 0) const;
  int find_first(const char*, size_t = 0) const;

  int find_last(char) const;
  int find_last(const char*) const;
  
  mystring left(size_t) const;
  mystring right(size_t) const;
  mystring sub(size_t, size_t) const;

  mystring lstrip() const;
  mystring rstrip() const;
  mystring strip() const;

  mystring operator+=(const char* that)
  {
    append(that, strlen(that));
    return *this;
  }

  mystring operator+=(const mystring& that)
  {
    append(that.rep->buf, that.rep->length);
    return *this;
  }
  
};

mystring operator+(const mystring&, const mystring&);
mystring operator+(const mystring&, const char*);

inline ostream& operator<<(ostream& out, const mystring& str)
{
  out.write(str.c_str(), str.length());
  return out;
}

istream& operator>>(istream& in, mystring& str);

typedef mystring string;

#endif
